"""
model
~~~~~

Common data elements for the yadr package.
"""
from collections import UserString
from enum import Enum, EnumMeta, auto
from typing import Generic, NamedTuple, Sequence, Union, Tuple, TypeVar


# YADN Tokens.
class Token(Enum):
    START = auto()
    END = auto()
    AS_OPERATOR = auto()
    BOOLEAN = auto()
    CHOICE_OPERATOR = auto()
    CHOICE_OPTIONS = auto()
    COMPARISON_OPERATOR = auto()
    DICE_OPERATOR = auto()
    EX_OPERATOR = auto()
    GROUP_OPEN = auto()
    GROUP_CLOSE = auto()
    KEY = auto()
    KV_DELIMITER = auto()
    MAP = auto()
    MAP_CLOSE = auto()
    MAP_END = auto()
    MAP_OPEN = auto()
    MAPPING_OPERATOR = auto()
    MD_OPERATOR = auto()
    MEMBER = auto()
    MEMBER_DELIMITER = auto()
    NAME = auto()
    NAME_DELIMITER = auto()
    NEGATIVE_SIGN = auto()
    NUMBER = auto()
    OPERATOR = auto()
    OPTIONS_OPERATOR = auto()
    PAIR_DELIMITER = auto()
    POOL = auto()
    POOL_CLOSE = auto()
    POOL_DEGEN_OPERATOR = auto()
    POOL_END = auto()
    POOL_GEN_OPERATOR = auto()
    POOL_OPEN = auto()
    POOL_OPERATOR = auto()
    QUALIFIER = auto()
    QUALIFIER_END = auto()
    QUALIFIER_DELIMITER = auto()
    ROLL_DELIMITER = auto()
    U_POOL_DEGEN_OPERATOR = auto()
    VALUE = auto()
    WHITESPACE = auto()


# Token categories.
op_tokens = (
    Token.CHOICE_OPERATOR,
    Token.COMPARISON_OPERATOR,
    Token.DICE_OPERATOR,
    Token.POOL_GEN_OPERATOR,
    Token.POOL_DEGEN_OPERATOR,
    Token.POOL_OPERATOR,
    Token.U_POOL_DEGEN_OPERATOR,
    Token.OPTIONS_OPERATOR,
    Token.OPERATOR,
    Token.AS_OPERATOR,
    Token.MD_OPERATOR,
    Token.EX_OPERATOR,
    Token.MAPPING_OPERATOR,
)
id_tokens = (
    Token.BOOLEAN,
    Token.NUMBER,
    Token.POOL,
    Token.QUALIFIER,
)

# Symbols for YADN tokens.
# This maps the symbols used in YADN to tokens for lexing. This isn't
# a direct mapping from the YADN specification document. It's just
# the basic things that can be handled by the lexer easily. More
# complicated things are handled through the lexer itself.
yadn_symbols_raw = {
    Token.START: '',
    Token.AS_OPERATOR: '+ -',
    Token.BOOLEAN: 'T F',
    Token.CHOICE_OPERATOR: '?',
    Token.CHOICE_OPTIONS: '',
    Token.COMPARISON_OPERATOR: '< > >= <= != ==',
    Token.DICE_OPERATOR: 'd d! dc dh dl dw',
    Token.EX_OPERATOR: '^',
    Token.GROUP_OPEN: '(',
    Token.GROUP_CLOSE: ')',
    Token.KEY: '',
    Token.KV_DELIMITER: ':',
    Token.MAP: '',
    Token.MAP_CLOSE: '}',
    Token.MAP_END: '',
    Token.MAP_OPEN: '{',
    Token.MAPPING_OPERATOR: 'm',
    Token.MD_OPERATOR: '* / %',
    Token.MEMBER_DELIMITER: ',',
    Token.NEGATIVE_SIGN: '-',
    Token.NAME: '',
    Token.NAME_DELIMITER: '=',
    Token.NUMBER: '0 1 2 3 4 5 6 7 8 9',
    Token.OPTIONS_OPERATOR: ':',
    Token.PAIR_DELIMITER: ',',
    Token.POOL_CLOSE: ']',
    Token.POOL: '',
    Token.POOL_END: '',
    Token.POOL_OPEN: '[',
    Token.POOL_DEGEN_OPERATOR: 'nb ns',
    Token.POOL_GEN_OPERATOR: 'g g!',
    Token.POOL_OPERATOR: 'pa pb pc pf ph pl pr p%',
    Token.QUALIFIER: '',
    Token.QUALIFIER_DELIMITER: '"',
    Token.QUALIFIER_END: '',
    Token.ROLL_DELIMITER: ';',
    Token.U_POOL_DEGEN_OPERATOR: 'C N S',
    Token.VALUE: '',
    Token.WHITESPACE: '',
}


# Classes.
class CompoundResult(Tuple):
    """The result of multiple rolls."""


# Types.
DiceMapping = dict[int, Union[int, str]]
NamedMap = Tuple[str, DiceMapping]
Pool = Tuple[int]
Result = Union[
    int,
    bool,
    str,
    Pool,
    NamedMap,
    None
]
TokenInfo = tuple[Token, Result]


# Symbols by token.
symbols = {k: v.split() for k, v in yadn_symbols_raw.items()}
symbols[Token.WHITESPACE] = [' ', '\t', '\n']



"""
maps
~~~~

A module for handling YADN dice maps.
"""
from typing import Callable, Optional

from yadr.base import BaseLexer
from yadr.model import NamedMap, Result, symbols, Token, TokenInfo


# Lexing.
class Lexer(BaseLexer):
    def __init__(self) -> None:
        state_map: dict[Token, Callable] = {
            Token.START: self._start,
            Token.END: self._start,
            Token.KV_DELIMITER: self._kv_delimiter,
            Token.MAP_CLOSE: self._map_close,
            Token.MAP_OPEN: self._map_open,
            Token.NAME_DELIMITER: self._name_delimiter,
            Token.NEGATIVE_SIGN: self._negative_sign,
            Token.NUMBER: self._number,
            Token.PAIR_DELIMITER: self._pair_delimiter,
            Token.QUALIFIER: self._qualifier,
            Token.QUALIFIER_END: self._qualifier_end,
            Token.WHITESPACE: self._whitespace,
        }
        symbol_map: dict[Token, list[str]] = symbols
        bracket_states: dict[Token, Token] = {
            Token.NEGATIVE_SIGN: Token.NUMBER,
            Token.QUALIFIER_DELIMITER: Token.QUALIFIER,
        }
        bracket_ends: dict[Token, Token] = {
            Token.QUALIFIER: Token.QUALIFIER_END,
        }
        result_map: dict[Token, Callable] = {
            Token.NUMBER: self._tf_number,
            Token.QUALIFIER: self._tf_qualifier,
        }
        no_store: list[Token] = [
            Token.START,
            Token.QUALIFIER_END,
            Token.WHITESPACE,
        ]
        init_state: Token = Token.START
        super().__init__(
            state_map,
            symbol_map,
            bracket_states,
            bracket_ends,
            result_map,
            no_store,
            init_state
        )

    # Result transformation rules.
    def _tf_number(self, value: str) -> int:
        return int(value)

    def _tf_qualifier(self, value: str) -> str:
        return value[1:-1]

    # Lexing rules.
    def _kv_delimiter(self, char: str) -> None:
        """Lex a key-value delimiter symbol."""
        can_follow = [
            Token.NEGATIVE_SIGN,
            Token.NUMBER,
            Token.QUALIFIER_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _map_close(self, char: str) -> None:
        """Lex a map close symbol."""
        can_follow: list[Token] = []
        self._check_char(char, can_follow)

    def _map_open(self, char: str) -> None:
        """Lex a map open symbol."""
        can_follow = [
            Token.MAP_CLOSE,
            Token.QUALIFIER_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _name_delimiter(self, char: str) -> None:
        """Lex a name delimiter symbol."""
        can_follow = [
            Token.NEGATIVE_SIGN,
            Token.NUMBER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _number(self, char: str) -> None:
        """Processing a number."""
        can_follow = [
            Token.KV_DELIMITER,
            Token.MAP_CLOSE,
            Token.PAIR_DELIMITER,
            Token.WHITESPACE,
        ]

        # Check here if the character is a digit because the checks in
        # Char are currently limited to tokens that no longer than two
        # characters. Check if the state is a number because white
        # space also ends up here, and we want white space to separate
        # numbers.
        if char.isdigit() and self.state == Token.NUMBER:
            self.buffer += char
        else:
            self._check_char(char, can_follow)

    def _negative_sign(self, char: str) -> None:
        """Processing a number."""
        can_follow = [
            Token.NUMBER,
        ]
        self._check_char(char, can_follow)

    def _pair_delimiter(self, char: str) -> None:
        """Lex a pair delimiter symbol."""
        can_follow = [
            Token.NEGATIVE_SIGN,
            Token.NUMBER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _qualifier(self, char: str) -> None:
        """Lex a qualifier."""
        self.buffer += char
        if self._is_token_start(Token.QUALIFIER_DELIMITER, char):
            new_state = Token.QUALIFIER_END
            self._change_state(new_state, char)

    def _qualifier_end(self, char: str) -> None:
        can_follow = [
            Token.MAP_CLOSE,
            Token.NAME_DELIMITER,
            Token.PAIR_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _start(self, char: str) -> None:
        """Initial lexer state."""
        if self.tokens:
            self.tokens = []
        can_follow = [
            Token.MAP_OPEN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)


# Parsing.
class Parser:
    def __init__(self):
        self.name = ''
        self.pairs = []
        self.buffer: Optional[int] = None
        self.state = Token.START
        self.state_map = {
            Token.START: self._start,
            Token.END: self._start,
            Token.KEY: self._key,
            Token.NAME: self._name,
            Token.VALUE: self._value,
        }

    def parse(self, tokens: tuple[TokenInfo, ...]) -> NamedMap:
        """Parse YADN dice mapping tokens."""
        for token_info in tokens:
            process = self.state_map[self.state]
            process(token_info)
        return (self.name, {k: v for k, v in self.pairs})

    # Parsing rules.
    def _key(self, token_info: tuple[Token, Result]) -> None:
        token, value = token_info
        if token == Token.NUMBER and isinstance(value, int):
            self.buffer = value
        elif token == Token.KV_DELIMITER:
            self.state = Token.VALUE
        elif token == Token.MAP_CLOSE:
            ...
        else:
            msg = f'KEY cannot contain {token.name}.'
            raise ValueError(msg)

    def _name(self, token_info: tuple[Token, Result]) -> None:
        token, value = token_info
        if token == Token.QUALIFIER:
            self.name = value
        elif token == Token.NAME_DELIMITER:
            self.state = Token.KEY
        elif token == Token.MAP_CLOSE:
            ...
        else:
            msg = f'NAME cannot contain {token.name}.'
            raise ValueError(msg)

    def _value(self, token_info: tuple[Token, Result]) -> None:
        token, value = token_info
        if (token == Token.QUALIFIER
                or token == Token.NUMBER and isinstance(value, int)):
            key = self.buffer
            pair = (key, value)
            self.pairs.append(pair)
            self.buffer = None
        elif token == Token.PAIR_DELIMITER:
            self.state = Token.KEY
        elif token == Token.MAP_CLOSE:
            ...
        else:
            msg = f'VALUE cannot contain {token.name}.'
            raise ValueError(msg)

    def _start(self, token_info: tuple[Token, Result]) -> None:
        token, value = token_info
        if token == Token.MAP_OPEN:
            self.state = Token.NAME
        else:
            msg = f'Dice mapping cannot start with a {value}'
            raise ValueError(msg)



"""
encode
~~~~~~

A Python object to YADN encoder.
"""
from yadr.model import CompoundResult, Result


class Encoder:
    def __init__(self, yadn: str = '') -> None:
        self.yadn = yadn

    # Public methods.
    def encode(self, data: Result | CompoundResult) -> str:
        """Turn computed Python object results into a YADN string."""
        if isinstance(data, bool):
            self._encode_bool(data)
        elif isinstance(data, int):
            self._encode_int(data)
        elif isinstance(data, str):
            self._encode_str(data)
        elif isinstance(data, CompoundResult):
            self._encode_compound_result(data)
        elif isinstance(data, tuple):
            self._encode_tuple(data)
        return self.yadn

    # Private methods.
    def _encode_bool(self, data: bool) -> None:
        if data:
            self.yadn = f'{self.yadn}T'
        else:
            self.yadn = f'{self.yadn}F'

    def _encode_compound_result(self, data: CompoundResult) -> None:
        for result in data[:-1]:
            self.encode(result)
            self.yadn += '; '
        else:
            self.encode(data[-1])

    def _encode_int(self, data: int) -> None:
        self.yadn = f'{self.yadn}{data}'

    def _encode_tuple(self, data: tuple) -> None:
        if isinstance(data[0], int):
            members = ', '.join(str(m) for m in data)
        else:
            quoted = [f'"{m}"' for m in data]
            members = ', '.join(str(m) for m in quoted)
        self.yadn = f'{self.yadn}[{members}]'

    def _encode_str(self, data: str) -> None:
        self.yadn = f'{self.yadn}"{data}"'



"""
pools
~~~~~

A module for handling YADN pools.
"""
from typing import Callable, Sequence

from yadr.base import BaseLexer
from yadr.model import symbols, Token, TokenInfo


class Lexer(BaseLexer):
    def __init__(self) -> None:
        state_map: dict[Token, Callable] = {
            Token.NUMBER: self._number,
            Token.MEMBER_DELIMITER: self._member_delimiter,
            Token.POOL: self._pool,
            Token.POOL_CLOSE: self._pool_close,
            Token.START: self._start,
            Token.WHITESPACE: self._whitespace,
            Token.END: self._start,
        }
        symbol_map: dict[Token, list[str]] = symbols
        bracket_states: dict[Token, Token] = {
            Token.NEGATIVE_SIGN: Token.NUMBER,
            Token.QUALIFIER_DELIMITER: Token.QUALIFIER,
            Token.POOL_OPEN: Token.POOL,
        }
        bracket_ends: dict[Token, Token] = {}
        result_map: dict[Token, Callable] = {
            Token.NUMBER: self._tf_number,
        }
        no_store: list[Token] = [
            Token.POOL_OPEN,
            Token.POOL_CLOSE,
            Token.WHITESPACE,
        ]
        super().__init__(
            state_map,
            symbol_map,
            bracket_states,
            bracket_ends,
            result_map,
            no_store,
            Token.START
        )
        self.process = self._start

    # Value transforms.
    def _tf_number(self, value: str) -> int:
        return int(value)

    # Lexing rules.
    def _member_delimiter(self, char: str) -> None:
        """Lex a member delimiter."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _number(self, char: str) -> None:
        """Lex a member."""
        can_follow = [
            Token.MEMBER_DELIMITER,
            Token.POOL_CLOSE,
            Token.WHITESPACE,
        ]

        # Check here if the character is a digit because the checks in
        # Char are currently limited to tokens that no longer than two
        # characters. Check if the state is a number because white
        # space also ends up here, and we want white space to separate
        # numbers.
        if char.isdigit() and self.state == Token.NUMBER:
            self.buffer += char
        else:
            self._check_char(char, can_follow)

    def _pool_close(self, char: str) -> None:
        """Lex a pool close."""
        msg = '[ cannot follow a ]'
        raise ValueError(msg)

    def _pool(self, char: str) -> None:
        """Lex a pool open."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _start(self, char: str) -> None:
        """Start lexing the string."""
        can_follow = [
            Token.POOL_OPEN,
        ]
        self._check_char(char, can_follow)


class Parser:
    def parse(self, tokens: Sequence[TokenInfo]) -> tuple[int, ...]:
        """Parse YADN pool tokens."""
        values = []
        for token in tokens:
            kind, value = token
            if kind == Token.NUMBER and isinstance(value, int):
                values.append(value)
            elif kind == Token.NUMBER:
                msg = f'NUMBER tokens should be ints. Found {type(value)}'
                raise TypeError(msg)
        return tuple(values)



"""
parse
~~~~~

Parse dice notation.
"""
from functools import wraps
import operator
from typing import Callable, Generic, Optional, Sequence, TypeVar

from yadr import operator as yo
from yadr.model import (
    CompoundResult,
    Result,
    Token,
    TokenInfo,
    op_tokens,
    id_tokens
)


# The dice map.
# This needs to not be a global value, but it will require a very large
# change to this module to get that to work. This will work for now.
dice_map: dict[str, dict] = {}


# Parser specific operations.
def map_result(result: int | tuple[int, ...],
               key: str) -> str | tuple[str, ...]:
    """Map a roll result to a dice map."""
    if isinstance(result, int):
        return dice_map[key][result]
    new_result = [map_result(n, key) for n in result]
    str_result = tuple(str(item) for item in new_result)
    return str_result


# Utility classes and functions.
class Tree:
    """A binary tree."""
    def __init__(self,
                 kind: Token,
                 value: Result,
                 left: Optional['Tree'] = None,
                 right: Optional['Tree'] = None,
                 dice_map: Optional[dict[str, dict]] = None) -> None:
        self.kind = kind
        self.value = value
        self.left = left
        self.right = right
        if dice_map is None:
            dice_map = {}
        self.dice_map = dice_map

    def __repr__(self):
        name = self.__class__.__name__
        return f'{name}(kind={self.kind}, value={self.value})'

    def compute(self):
        if self.kind in id_tokens:
            return self.value
        left = self.left.compute()
        right = self.right.compute()
        if self.kind in op_tokens:
            ops_by_symbol = yo.ops_by_symbol
            ops_by_symbol['m'] = self._map_result
            op = ops_by_symbol[self.value]
        else:
            msg = f'Unknown token {self.kind}'
            raise TypeError(msg)
        return op(left, right)

    def _map_result(self, result: int | tuple[int, ...],
                    key: str) -> str | tuple[str, ...]:
        """Map a roll result to a dice map."""
        if isinstance(result, int):
            return self.dice_map[key][result]
        new_result = [self._map_result(n, key) for n in result]
        str_result = tuple(str(item) for item in new_result)
        return str_result


class Unary(Tree):
    """A unary tree."""
    def __init__(self,
                 kind: Token,
                 value: Result,
                 child: Optional['Tree'] = None) -> None:
        self.kind = kind
        self.value = value
        self.child = child

    def compute(self):
        if self.kind in id_tokens:
            return self.value
        child = self.child.compute()
        if self.kind in op_tokens:
            op = yo.ops_by_symbol[self.value]
        return op(child)


# Parser class.
class Parser:
    def __init__(self) -> None:
        self.dice_map: dict[str, dict] = dict()
        self.top_rule = self._map_operator

    # Public method.
    def parse(self, tokens: Sequence[TokenInfo]) -> Result | CompoundResult:
        if (Token.ROLL_DELIMITER, ';') not in tokens:
            return self._parse_roll(tokens)        # type: ignore

        rolls = []
        while (Token.ROLL_DELIMITER, ';') in tokens:
            index = tokens.index((Token.ROLL_DELIMITER, ';'))
            roll = tokens[0:index]
            rolls.append(roll)
            tokens = tokens[index + 1:]
        else:
            rolls.append(tokens)
        results: Sequence[Result] = []
        for roll in rolls:
            results.append(self.parse(roll))       # type: ignore
            results = [result for result in results if result is not None]
        if len(results) > 1:
            return CompoundResult(results)
        elif results:
            return results[0]
        return None

    def _parse_roll(self, tokens: Sequence[TokenInfo]) -> Result:
        """Parse a sequence of YADN tokens."""
        def make_tree(kind, value):
            return Tree(kind, value, dice_map=self.dice_map)

        trees = [make_tree(kind, value) for kind, value in tokens]
        trees = trees[::-1]
        parsed = self.top_rule(trees)
        if parsed:
            return parsed.compute()
        return None

    # Parsing rules.
    def _identity(self, trees: list[Tree]) -> Tree | None:
        """Parse an identity."""
        identity_tokens = [
            Token.BOOLEAN,
            Token.NUMBER,
            Token.POOL,
            Token.QUALIFIER,
        ]
        tree = trees.pop()
        if tree.kind in identity_tokens:
            return tree
        elif tree.kind == Token.MAP:
            name, map_ = tree.value                          # type: ignore
            self.dice_map[name] = map_
            return None
        elif tree.kind == Token.GROUP_OPEN:
            expression = self.top_rule(trees)
            if trees[-1].kind == Token.GROUP_CLOSE:
                _ = trees.pop()
            return expression
        else:
            msg = f'Unrecognized token {tree.kind}'
            raise TypeError(msg)

    def _pool_gen_operator(self, trees: list[Tree]) -> Tree:
        """Parse pool generation operator."""
        rule = self._binary_operator
        rule_affects = Token.POOL_GEN_OPERATOR
        next_rule = self._identity
        return rule(rule_affects, next_rule, trees)

    def _pool_operator(self, trees: list[Tree]) -> Tree:
        """Parse pool operator."""
        rule = self._binary_operator
        rule_affects = Token.POOL_OPERATOR
        next_rule = self._pool_gen_operator
        return rule(rule_affects, next_rule, trees)

    def _u_pool_degen_operator(self, trees: list[Tree]) -> Tree:
        """Parse unary pool degeneration."""
        rule = self._unary_operator
        rule_affects = Token.U_POOL_DEGEN_OPERATOR
        next_rule = self._pool_operator
        return rule(rule_affects, next_rule, trees)

    def _pool_degen_operator(self, trees: list[Tree]) -> Tree:
        """Parse unary pool degeneration."""
        rule = self._binary_operator
        rule_affects = Token.POOL_DEGEN_OPERATOR
        next_rule = self._u_pool_degen_operator
        return rule(rule_affects, next_rule, trees)

    def _dice_operator(self, trees: list[Tree]) -> Tree:
        """Parse dice operators."""
        rule = self._binary_operator
        rule_affects = Token.DICE_OPERATOR
        next_rule = self._pool_degen_operator
        return rule(rule_affects, next_rule, trees)

    def _ex_operator(self, trees: list[Tree]) -> Tree:
        """Parse exponentiation."""
        rule = self._binary_operator
        rule_affects = Token.EX_OPERATOR
        next_rule = self._dice_operator
        return rule(rule_affects, next_rule, trees)

    def _md_operator(self, trees: list[Tree]) -> Tree:
        """Parse addition and subtraction."""
        rule = self._binary_operator
        rule_affects = Token.MD_OPERATOR
        next_rule = self._ex_operator
        return rule(rule_affects, next_rule, trees)

    def _as_operator(self, trees: list[Tree]) -> Tree:
        """Parse addition and subtraction."""
        rule = self._binary_operator
        rule_affects = Token.AS_OPERATOR
        next_rule = self._md_operator
        return rule(rule_affects, next_rule, trees)

    def _comparison_operator(self, trees: list[Tree]) -> Tree:
        """Parse comparisons."""
        rule = self._binary_operator
        rule_affects = Token.COMPARISON_OPERATOR
        next_rule = self._as_operator
        return rule(rule_affects, next_rule, trees)

    def _options_operator(self, trees: list[Tree]) -> Tree:
        """Parse comparisons."""
        rule = self._binary_operator
        rule_affects = Token.OPTIONS_OPERATOR
        next_rule = self._comparison_operator
        return rule(rule_affects, next_rule, trees)

    def _choice_operator(self, trees: list[Tree]) -> Tree:
        """Parse chocies."""
        rule = self._binary_operator
        rule_affects = Token.CHOICE_OPERATOR
        next_rule = self._options_operator
        return rule(rule_affects, next_rule, trees)

    def _map_operator(self, trees: list[Tree]) -> Tree:
        """Parse dice mapping operators."""
        rule = self._binary_operator
        rule_affects = Token.MAPPING_OPERATOR
        next_rule = self._choice_operator
        return rule(rule_affects, next_rule, trees)

    # Base rules.
    def _binary_operator(self, rule_affects: Token,
                         next_rule: Callable,
                         trees: list[Tree]) -> Tree:
        """Parse a binary operator."""
        left = next_rule(trees)
        while trees and trees[-1].kind == rule_affects:
            tree = trees.pop()
            tree.left = left
            tree.right = next_rule(trees)
            left = tree
        return left

    def _unary_operator(self, rule_affects: Token,
                        next_rule: Callable,
                        trees: list[Tree]) -> Tree:
        """Parse an unary operator."""
        if trees[-1].kind == rule_affects:
            tree = trees.pop()
            unary = Unary(tree.kind, tree.value)
            unary.child = next_rule(trees)
            return unary
        return next_rule(trees)
